package it.polito.elite.visualization.ontoviewer.models;

import it.polito.elite.ontology.NodeTypeAbstraction;
import it.polito.elite.ontology.OWLOntologyWrapper;
import it.polito.elite.ontology.OntologyAbstraction;
import it.polito.elite.visualization.ontoviewer.enums.LinkDirectionEnum;
import it.polito.elite.visualization.ontoviewer.enums.NodeTypesEnum;
import it.polito.elite.visualization.ontoviewer.enums.QueryLanguageEnum;
import it.polito.elite.visualization.ontoviewer.tools.Relation3DObject;
import it.polito.elite.visualization.ontoviewer.userInterfaces.VisualizationOptions;

import java.awt.Color;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;

import javax.swing.JTree;
import javax.swing.tree.DefaultMutableTreeNode;

import edu.stanford.smi.protege.model.KnowledgeBase;
import edu.stanford.smi.protegex.owl.model.OWLModel;

/**
 * Used to enrich the default data model built by the ontology abstraction
 * faclity, that relies on API like prot�g� or jena. In particular DataModel
 * class retrieves the incoming relations (between concepts and between
 * instances)
 * 
 * @author Alessio
 * 
 */
public class KnowledgeBaseModel {
	
	private OntologyAbstraction onto;
	
	private JTree tree;

	private HashSet<String> roots = null;

	private Hashtable<String, ConceptData> concepts = null;

	private Hashtable<String, InstanceData> instances = null;
	
	private Hashtable<String, HashSet<String>> logicalViews= null;
	
	static private String logicViewAnnotationURI="belongsToView";
	static private String coloredItemAnnotationURI="hasColor";
	
	public KnowledgeBaseModel(KnowledgeBase kb) throws Exception {
		concepts = new Hashtable<String, ConceptData>();
		instances = new Hashtable<String, InstanceData>();
		roots = new HashSet<String>();
		// creates a wrapper 4 querying the KB, used to building the internal model
		onto = OntologyAbstraction.createOntologyAbstraction(kb);
		//creates a structure handling the logical views defined in the ontology
		logicalViews = new Hashtable<String, HashSet<String>>();
		
		this.loadRelationsData();
		this.tree = this.loadNodes();
		this.retrieveIncomingLinks();
	}
	
	private void loadRelationsData(){
		Hashtable<String, RelationType> relations=this.onto.getAllRelations(false);
		/*
		 * Checks if relations have a custom color associated (by means of an annotation property)
		 * and if this is true loads the color and store the info
		 */
		for(String relURI: relations.keySet()){
			Hashtable<String,Object> relAnnotations= this.onto.getAnnotationProperties(relURI, false);
			for(String annotationURI: relAnnotations.keySet()){
				if(annotationURI.equals(VisualizationOptions.s_LogicViewsOntologyPrefix+":"+KnowledgeBaseModel.coloredItemAnnotationURI)){
					String colorAsString = (String) relAnnotations.get(annotationURI);
					String[] colorComponents = colorAsString.split(" ");
					if(colorComponents.length == 3){
						int r=Integer.parseInt(colorComponents[0]);
						int g=Integer.parseInt(colorComponents[1]);
						int b=Integer.parseInt(colorComponents[2]);
						Color c=new Color(r,g,b);
						Relation3DObject.setColorToRelation(c, relURI);
					}
				}
			}
		}
	}
	
	/**
	 * 
	 * @param onto
	 * @return
	 */
	private JTree loadNodes() {
		// Build the JTree and build up the internal Data Model
		DefaultMutableTreeNode root = new DefaultMutableTreeNode("owl:Thing");
		this.roots = onto.getRoots(false);

		for (String uri : roots) {
			deepFirstTreeVisit(onto, uri, 1, null, root);
		}

		return new JTree(root);
	}
	
	private void cecksConceptMembershipToViews(String uri){
		Hashtable<String,Object> table=onto.getAnnotationProperties(uri, true);
		for(String anno: table.keySet() ){
			//if the annotation property is about logic views "belongsToView" store the info
			if(anno.equals(VisualizationOptions.s_LogicViewsOntologyPrefix+":"+KnowledgeBaseModel.logicViewAnnotationURI)){
				String annotationValue=table.get(anno).toString();
				
				HashSet<String> viewElements=null;
				if(this.logicalViews.keySet().contains(annotationValue)){
					viewElements=this.logicalViews.get(annotationValue);
				}
				else{	
					viewElements=new HashSet<String>();
					this.logicalViews.put(annotationValue, viewElements);
				}	
				viewElements.add(uri);
			}
		}
	}
	
	/**
	 * Recursive exploration procedure triggered by loadNodes
	 */
	private ConceptData deepFirstTreeVisit(OntologyAbstraction onto, String uri,
			int lvl, String parent, DefaultMutableTreeNode parentNode) {

//belongs to some scene?		
		cecksConceptMembershipToViews(uri);
		
		if (this.concepts.containsKey(uri)) {
			// Uri are uniques, so the concept is already present in some hierarchy
			ConceptData cd = this.concepts.get(uri);
			cd.addAdoptiveParent(parent);
			return cd;
		}
		// Creating the jtree node
		DefaultMutableTreeNode node = new DefaultMutableTreeNode(uri);
		parentNode.add(node);
		// Setting the concept data
		ConceptData concept = new ConceptData(uri);
		concept.setParent(parent);
		concept.setNode(node);

		// Retrieving & Setting outgoing and datatype properties
		concept.setOutgoingLinks(onto.getObjectProperties(uri, true, false));
		concept.setDatatypeLinks(onto.getDatatypeProperties(uri));
		//Retrieving and setting restrictions 4 datatype and obj properties, then compacting
		this.addRestrictionsToOutgoingRelations(uri, concept);
		this.compactOutgoingRelations(concept);
		
		//Setting intesections, unions, oneOf..
		this.onto.setLogicalConstraints(uri, concept);
		
		// Retrieving & Setting instances
		concept.setInstancesURI(onto.getInstancesOfConcept(uri));
		for (String instance : concept.getInstancesURI()) {
			InstanceData idata = new InstanceData(instance);
			idata.setPertainingConceptURI(uri);
			idata.addOutgoingLinks(onto.getInstancesObjectProperties(instance));
			idata.addDatatypeLinks(onto.getInstancesDatatypeProperties(instance));
			this.onto.setLogicalConstraints(instance, idata);
			
			this.instances.put(instance, idata);
		}

		// Setting Children
		boolean childrenHasInstances=false;
		HashSet<String> childrenSet = onto.getSubClasses(uri);
		for (String subUri : childrenSet) {
			ConceptData c1 = deepFirstTreeVisit(onto, subUri, lvl + 1, uri, node);
			concept.addChild(subUri, c1.getNumberOfChildren());
			if (c1.hasInstances() || c1.hasNonDirectInstances())
				childrenHasInstances=true;
		}
		if(childrenHasInstances)
			concept.setNonDirectInstances(true);
		this.concepts.put(uri, concept);
		return concept;
	}
	
	private void compactOutgoingRelations(ConceptData concept){
		Hashtable<String, HashSet<RelationData>> objProps=concept.getOutgoingLinks();	
		for(String relURI: objProps.keySet()){
			HashSet<RelationData>relData = objProps.get(relURI);
			HashSet<RelationData>compact=new HashSet<RelationData>(relData.size());
			
			for(Iterator<RelationData> itOriginal=relData.iterator();itOriginal.hasNext();){
				RelationData original=itOriginal.next();
				HashSet<RelationData> toRemove=new HashSet<RelationData>(relData.size());
				for(RelationData comparison:relData){
					if( !original.equals(comparison)  && original.hasSameRelatedNode(comparison)){	
						if(comparison.isCardinalityConstrained()){
							original.setFixedCardinality(comparison.getFixedCardinality());
							original.setMaxCardinality(comparison.getMaxCardinality());
							original.setMinCardinality(comparison.getMinCardinality());
						}
						toRemove.add(comparison);
					}
				}
				itOriginal.remove();
				compact.add(original);
				relData.removeAll(toRemove);
				itOriginal=relData.iterator();
			}
			objProps.put(relURI, compact);
		}
	}
	
	/**
	 * Adds the logical restrictions specified for a given concept to its object properties.
	 * it simply adds them without any check. The model is considered correct (it doesn't check
	 * conflicts between ALL values FROM and SOME values FROM or HAS VALUE)
	 * 
	 * @param uri The Uri of the concept
	 * @param concept Its ConceptData
	 */
	private void addRestrictionsToOutgoingRelations(String uri, ConceptData concept){
		
//		getting the restrictions
		Hashtable<String, HashSet<NodeTypeAbstraction>> restrictions = this.onto.getRestrictions(uri);
		for(String relURI: restrictions.keySet()){
			RelationType rType = this.onto.getRelationType(relURI);
			//Getting the relation ranges set if already present or creating a new one for 
			//the ranges inferred from restrictions 
			HashSet<RelationData> setOfRelationData;
			if(concept.getOutgoingLinks().keySet().contains(relURI))
				setOfRelationData = new HashSet<RelationData>( concept.getOutgoingLinks().get(relURI) );
			else if(concept.getDatatypeLinks().keySet().contains(relURI)){
				setOfRelationData = new HashSet<RelationData>( concept.getDatatypeLinks().get(relURI) );
			}
			else{
				setOfRelationData = new HashSet<RelationData>();
			}
			
			HashSet<NodeTypeAbstraction> uriRestrictions=restrictions.get(relURI);
			//Creating the RelationData entry/entries
			for(NodeTypeAbstraction myRestriction: uriRestrictions){
							
				if( myRestriction.getType() == NodeTypesEnum.ExistentialQuantifier 		||
					myRestriction.getType() == NodeTypesEnum.QualifiedCardinality 		||	
					myRestriction.getType() == NodeTypesEnum.QualifiedMinCardinality 	||
					myRestriction.getType() == NodeTypesEnum.QualifiedMaxCardinality 		){
					
					NodeTypeAbstraction endNode=myRestriction.getRestrictionFiller();
					//Deleting preexisting RelationData (if simple) towards the same endnode of the restriction
					for(Iterator<RelationData> it=setOfRelationData.iterator();it.hasNext();){
						RelationData myRData=it.next();
						NodeTypeAbstraction myRData_endNode=myRData.getRelatedNode();
						if(endNode.toString().equals(myRData_endNode.toString()))
							it.remove();
					}
					//Create the RelationData for the restriction
					RelationData rdata = new RelationData(relURI, LinkDirectionEnum.out, rType); 
					rdata.setRelatedNode(endNode);
					//Setting cardinality
					if(myRestriction.getType() == NodeTypesEnum.QualifiedCardinality){
						rdata.setFixedCardinality(myRestriction.getCardinality());
					}
					else if(myRestriction.getType() == NodeTypesEnum.QualifiedMinCardinality){
						//If there is a preexisting MAX cardinality or Existential Quantifier -> adding the MIN constraint
						boolean found=false;
						for(Iterator<RelationData> it=setOfRelationData.iterator();it.hasNext();){
							RelationData myRData=it.next();
							NodeTypeAbstraction myRData_endNode=myRData.getRelatedNode();
							if(endNode.toString().equals(myRData_endNode.toString()) &&
									endNode.getType() == NodeTypesEnum.QualifiedMaxCardinality){
								myRData.setMinCardinality(myRestriction.getCardinality());
								found=true;
							}
						}
						if(found) 
							continue;
						else
							rdata.setMinCardinality(myRestriction.getCardinality());
					}
					else if(myRestriction.getType() == NodeTypesEnum.QualifiedMaxCardinality){
//						If there is a preexisting MIN cardinality or Existential Quantifier -> adding the MAX constraint
						boolean found=false;
						for(Iterator<RelationData> it=setOfRelationData.iterator();it.hasNext();){
							RelationData myRData=it.next();
							NodeTypeAbstraction myRData_endNode=myRData.getRelatedNode();
							if(endNode.toString().equals(myRData_endNode.toString()) &&
									endNode.getType() == NodeTypesEnum.QualifiedMinCardinality&&
									endNode.getType() == NodeTypesEnum.ExistentialQuantifier){
								myRData.setMaxCardinality(myRestriction.getCardinality());
								found=true;
							}
						}
						if(found) 
							continue;
						else
							rdata.setMaxCardinality(myRestriction.getCardinality());
					}
					else{//Existential Quantifier
						rdata.setMinCardinality(1);
					}
					setOfRelationData.add(rdata);
				}
				else if( myRestriction.getType() == NodeTypesEnum.Cardinality 		||	
						myRestriction.getType() == NodeTypesEnum.MinCardinality 	||
						myRestriction.getType() == NodeTypesEnum.MaxCardinality ){
//					Create the RelationData for the restriction
					NodeTypeAbstraction endNode=this.onto.getPropertyRange(relURI);
					if(endNode != null){
						RelationData rdata = new RelationData(relURI, LinkDirectionEnum.out, rType); 
						rdata.setRelatedNode(endNode);
	//					Setting cardinality
						if(myRestriction.getType() == NodeTypesEnum.Cardinality){
							rdata.setFixedCardinality(myRestriction.getCardinality());
						}
						else if(myRestriction.getType() == NodeTypesEnum.MinCardinality){
							rdata.setMinCardinality(myRestriction.getCardinality());
						}
						else if(myRestriction.getType() == NodeTypesEnum.MaxCardinality){
							rdata.setMaxCardinality(myRestriction.getCardinality());
						}
						setOfRelationData.add(rdata);
					}
				}
					else if(myRestriction.getType() == NodeTypesEnum.HasValue){
						//ranges.clear();
						RelationData rdata = new RelationData(relURI, LinkDirectionEnum.out, rType);
						rdata.setRelatedNode(myRestriction.getRestrictionFiller());
						rdata.setHasValue(true);
						setOfRelationData.add(rdata);
					}
					else if (myRestriction.getType() == NodeTypesEnum.UniversalQuantifier){
						RelationData rdata = new RelationData(relURI, LinkDirectionEnum.out, rType);
						rdata.setRelatedNode(myRestriction.getRestrictionFiller());
						rdata.setUniversal(true);
						setOfRelationData.add(rdata);
					}
			}
			if(rType.isDatatype())
				concept.datatypeLinks.put(relURI, setOfRelationData);
			else
				concept.outgoingLinks.put(relURI, setOfRelationData);
		}
	}
	
	/**
	 * Infers & Set the incoming ObjectProperties of every node from the
	 * outgoing ones
	 */
	private void retrieveIncomingObjProperties() {
		for (String uri : this.concepts.keySet()) {
			
			ConceptData c = this.concepts.get(uri);
			HashSet<String> toRemove=new HashSet<String>();
			for (String relURI : c.getOutgoingLinks().keySet()) {
				HashSet<RelationData> set = c.getOutgoingLinks().get(relURI);
				if(set == null) continue;
				//try{
				for(Iterator<RelationData> it=set.iterator();it.hasNext();){
					RelationData myRelationData=it.next();
					NodeTypeAbstraction myRange=myRelationData.getRelatedNode();
					if(myRange!=null && myRange.getType()==NodeTypesEnum.NamedClass){
						String endUri=myRange.toString();
						if(uri.equals(endUri)){//SelfLoop
							if(set.size()>1)
								it.remove();
							else
								toRemove.add(relURI);
							c.addSelfLoopLink(myRelationData);
						}
						else{
							ConceptData csource = this.concepts.get(endUri);
							NodeTypeAbstraction domain=this.onto.createNodeTypeAbstraction(
									NodeTypesEnum.NamedClass, uri);
							String incomingRelURI = myRelationData.getRelationName();
							RelationType rt = this.onto.getRelationType(incomingRelURI);
							RelationData incoming=new RelationData(incomingRelURI,
									myRelationData.getDir(), rt);
							incoming.setRelatedNode(domain);
							if(myRelationData.isCardinalityConstrained()){
								incoming.setFixedCardinality(myRelationData.getFixedCardinality());
								incoming.setMinCardinality(myRelationData.getMinCardinality());
								incoming.setMaxCardinality(myRelationData.getMaxCardinality());
							}
							if(domain!=null)
								csource.addIncomingLink(relURI, incoming);
						}
					}
				}
				//}catch(Exception e){	System.out.println(uri+" "+relURI+" "+set.size());}
			}
			for(String relURI:toRemove){
				c.getOutgoingLinks().remove(relURI);
			}
		}
	}

	/**
	 * Infers & Set the incoming facts of every instances from the outgoing ones
	 */
	private void retrieveIncomingFacts() {
		for (String instanceURI : this.instances.keySet()) {
			
			InstanceData idata = this.instances.get(instanceURI);
			HashSet<String> toRemove=new HashSet<String>();
			for (String relURI : idata.getOutgoingLinks().keySet()) {
				HashSet<NodeTypeAbstraction> set = idata.getOutgoingLinks().get(relURI);
				
				for(NodeTypeAbstraction myRange: set){
					if(myRange.getType()== NodeTypesEnum.NamedInstance){
						
						String endUri=myRange.toString();
						if(instanceURI.equals(endUri)){
							if(set.size()>1)
								set.remove(myRange);
							else
								toRemove.add(relURI);
							idata.addSelfLoopLink(relURI);
						}
						else{
							InstanceData isource = this.instances.get(endUri);
							NodeTypeAbstraction domain=this.onto.createNodeTypeAbstraction(
									NodeTypesEnum.NamedInstance, instanceURI);
							if(domain!=null)
								isource.addIncomingLink(relURI, domain);
						}
						
					}
				}
				
			}
			for(String relURI:toRemove){
				idata.getOutgoingLinks().remove(relURI);
			}
		}
	}



	public Set<String> retrieveAllInstancesOf(String uri) {
		ConceptData cd = this.getConceptData(uri);
		HashSet<String> set = new HashSet<String>();
		set.addAll(cd.getInstancesURI());
		for (String s : cd.getChildren()) {
			set.addAll(this.retrieveAllInstancesOf(s));
		}
		return set;
	}

	private void retrieveIncomingLinks() {
		this.retrieveIncomingObjProperties();
		this.retrieveIncomingFacts();
	}

	public ConceptData getConceptData(String conceptUri) {
		return this.concepts.get(conceptUri);
	}

	public InstanceData getInstanceData(String instanceUri) {
		return this.instances.get(instanceUri);
	}

	public HashSet<String> getRoots() {
		return roots;
	}

	public int getRootsNumber() {
		return roots.size();
	}

	public Set<String> getViews(){
		return this.logicalViews.keySet();
	}
	
	public HashSet<String> getViewElements(String viewID){
		return this.logicalViews.get(viewID);
	}
	
	public void saveViewElements(String viewID, HashSet<String> concepts){
		this.logicalViews.put(viewID, concepts);
		for(String uri: concepts){
			this.onto.addAnnotationProperty4URI(uri, VisualizationOptions.s_LogicViewsOntologyPrefix+
					":"+KnowledgeBaseModel.logicViewAnnotationURI, viewID);
		}
	}
	
	public void deleteView(String viewID){
		HashSet<String> toRemove=this.logicalViews.remove(viewID);
		for(String uri: toRemove){
			this.onto.removeAnnotationProperty4URI(uri, VisualizationOptions.s_LogicViewsOntologyPrefix+
					":"+KnowledgeBaseModel.logicViewAnnotationURI, viewID);
		}
	}
	
	public void setColorToRelation(String relURI, Color col){
		Relation3DObject.setColorToRelation(col, relURI);
		String colorAsStr=new String(col.getRed()+" "+col.getGreen()+" "+col.getBlue());
		this.onto.addAnnotationProperty4URI(relURI, VisualizationOptions.s_LogicViewsOntologyPrefix+
				":"+KnowledgeBaseModel.coloredItemAnnotationURI, colorAsStr);	
	}
	
	public void removeColorToRelation(String relURI){
		Color col=Relation3DObject.getRelationColor(relURI);
		Relation3DObject.removeRelationColor(relURI);
		String colorAsStr=new String(col.getRed()+" "+col.getGreen()+" "+col.getBlue());
		this.onto.removeAnnotationProperty4URI(relURI, VisualizationOptions.s_LogicViewsOntologyPrefix+
				":"+KnowledgeBaseModel.coloredItemAnnotationURI, colorAsStr);	
		
	}
	
	public JTree getTree() {
		return tree;
	}

	public void confirmLogicViewSupport(){
		onto.confirmLogicViewSupport();
	}
	
	public HashSet <NodeTypeAbstraction> query (String queryText, QueryLanguageEnum queryLang, int queryType)
	{
		System.out.println("Received query: "+queryText);
		if(this.onto.canQuery())
		{
			System.out.println("Can query: TRUE");
			if(this.onto instanceof OWLOntologyWrapper)
			{
				System.out.println("IsOWL: TRUE");
				return ((OWLOntologyWrapper)this.onto).query(queryText, queryLang, queryType);
			}
		}
		return null;
	}
	
	public OWLModel getOWLModel()
	{
		if(this.onto instanceof OWLOntologyWrapper)
		{
			OWLOntologyWrapper owlOnto = (OWLOntologyWrapper) this.onto;
			return owlOnto.getModel();
		}
		return null;
	}
	
	public Hashtable<String, RelationType> getRelationTypes(){
		return this.onto.getAllRelations(false);
	}
	
}
