package it.polito.elite.visualization.ontoviewer.tools;

import java.awt.Font;

import javax.media.j3d.Appearance;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.ColoringAttributes;
import javax.media.j3d.Font3D;
import javax.media.j3d.FontExtrusion;
import javax.media.j3d.Material;
import javax.media.j3d.OrientedShape3D;
import javax.media.j3d.Shape3D;
import javax.media.j3d.Text3D;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.media.j3d.TransparencyAttributes;
import javax.vecmath.Color3f;
import javax.vecmath.Point3f;

import com.sun.j3d.utils.geometry.Box;

/**
 * This class offers an utility to draw cubes as a building block
 * for scenes. It exposes a dummy constructor, a group of set methods
 * in order to explicitly set characteristics as transparency, the radius..
 * It provides also a  method createColoredCube which return a java 3D TransformGroup 
 * that will be included by the scene itself.
 * An example of its use is:
 * 		TransformGroup tgx = new TransformGroup();
 * 		String label="...";
 * 		InstanceCube cube = new InstanceCube();
 *		tgx.addChild(cube.createColoredCube(......));	
 *
 * @author alessio.bosca@polito.it
 *
 */
public class InstanceCube {


	// used as the UserData associated to the 3d obj on the screen
	private String cubeLabel = null;

	private final Color3f black = new Color3f(.0f, .0f, .0f);
	
	//default label color is black
	private Color3f labelColor=black;
	// sphere base color
	private Color3f cubeColor;
	//scale factor to render on the scene the label that is originally
	//created using a pointsize of 2 (see awt font & java3D font3D classes for details)
	private float labelScale=.018f;
	//The font family uset for writing the label
	private String fontFamily="Dialog";
	//The font modifier used for writing the label
	private int fontModifier=Font.PLAIN;
	// transparency of the sphere material. Default opaque.
	private float transparency = .0f;
	// Cube Side
	private float side = .04f;

	// the cube drawn
	private Box cube = null;
	
	private boolean isWithLabel=true;
	
	public void setIsWithLabel(boolean isWithLabel) {
		this.isWithLabel = isWithLabel;
	}

	/**
	 * Void constructor. It does notingh. To draw relation, one of the
	 * createXXXSphere function must be used
	 */
	public InstanceCube() {	}
	
	/**
	 * Creates and adds a Sphere to a TransformGroup. The user data (which is 
	 * the information retrieved by the Picker Behavior) of the 3D sphere object 
	 * is set as the label string. 
	 * @param color The Color of the Sphere
	 * @param label The label of the sphere (can be null)
	 * @param isPickable Indicates if the sphere should be pickable by user.
	 * @return the TransformGroup containing the sphere
	 */
	public BranchGroup createColoredCube(Color3f color, String label, boolean hasPertainingFacts) {
		this.cubeColor = color;
		this.cubeLabel = label;
		return this.drawInstanceCube(hasPertainingFacts);
	}

	/**
	 * @param transparency  The transparency to set.
	 */
	public void setTransparency(float transparency) {
		this.transparency = transparency;
	}
	
	
	/**
	 * This method actually draws the sphere
	 * @return the TransformGroup containing the sphere and the label (if present)
	 */
	private BranchGroup drawInstanceCube(boolean hasFacts) {
		// crea la sfera e la aggiunge al tgx
		BranchGroup bgx = new BranchGroup();
		Material material = new Material(this.cubeColor, black, this.cubeColor,
				this.cubeColor, 10.0f);
		
		material.setLightingEnable(true);
		Appearance app = new Appearance();
		app.setCapability(Appearance.ALLOW_COLORING_ATTRIBUTES_WRITE);
		app.setCapability(Appearance.ALLOW_MATERIAL_WRITE);
		app.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_WRITE);
		if (this.transparency > 0) {
			TransparencyAttributes ta = new TransparencyAttributes();
			ta.setTransparency(this.transparency);
			ta.setTransparencyMode(TransparencyAttributes.BLENDED);
			app.setTransparencyAttributes(ta);
		}
		app.setMaterial(material);
		
		cube = new Box(side, side, side, Box.ENABLE_APPEARANCE_MODIFY
				| Box.GENERATE_NORMALS, app);

		
		cube.setCapability(Box.ALLOW_BOUNDS_READ);
		cube.setCapability(Shape3D.ALLOW_LOCAL_TO_VWORLD_READ);
			
		if(cubeLabel!=null){
			cube.setUserData(new String(cubeLabel));
			if(hasFacts)
				cube.setPickable(true);
			if(this.isWithLabel)
				bgx.addChild(this.addCubeLabel());
		}
		else
			cube.setPickable(false);
		
		bgx.addChild(cube);
		
		return bgx;
	}
	
	
	
	/**
	 * This method creates a TransformGroup containing the label.
	 * @return the TransformGroup containing the label
	 */
	private TransformGroup addCubeLabel() {
		// inserisce la label e la aggiunge al tgx
		String textString = cubeLabel.substring(cubeLabel.indexOf("#") + 1);
		float sl = textString.length();

		// create 3D text
		Font3D f3d = new Font3D(new Font(this.fontFamily, this.fontModifier, 2),
				new FontExtrusion());

		Point3f textPt;
		float fontHeight=(this.side+.02f)/this.labelScale;
		textPt = new Point3f(-sl / 2.0f, fontHeight, 0.0f);
		Text3D txt = new Text3D(f3d, textString, textPt);

		OrientedShape3D textShape = new OrientedShape3D();
		textShape.setGeometry(txt);
		textShape.setPickable(false);

		TransformGroup objScale = new TransformGroup();
		Transform3D textMat = new Transform3D();
		textMat.setScale(this.labelScale);
		objScale.setTransform(textMat);

		Appearance apText = new Appearance();
		ColoringAttributes colorTxt = new ColoringAttributes(labelColor, 1);
		apText.setColoringAttributes(colorTxt);
		apText.setCapability(Appearance.ALLOW_MATERIAL_WRITE);
		textShape.setAppearance(apText);

		textShape.setAlignmentMode(OrientedShape3D.ROTATE_ABOUT_POINT);
		Point3f rotationPt = new Point3f(0.0f,0f, 0.0f);
		textShape.setRotationPoint(rotationPt);

		objScale.addChild(textShape);
		return objScale;
	}
	
	
	
	
	/**
	 * Sets color and dimension scale for the label 
	 * @param labelColor
	 * @param scale
	 */
	public void setLabelAppearance(Color3f labelColor,float scale) {
		this.labelColor = labelColor;
		this.labelScale = scale;
	}

	/**
	 * @param fontFamily The fontFamily to set.
	 */
	public void setFontFamily(String fontFamily) {
		this.fontFamily = fontFamily;
	}

	/**
	 * @param fontModifier The fontModifier to set.
	 */
	public void setFontModifier(int fontModifier) {
		this.fontModifier = fontModifier;
	}

	/**
	 * @return Returns the side.
	 */
	public float getSide() {
		return side;
	}

	/**
	 * @param side The side to set.
	 */
	public void setSide(float side) {
		this.side = side;
	}

	
}

